"use strict";

var logger, ffmpeg;

var cfg = require('../configuration.js');
var fs = require('fs');
var FirstN = require('first-n-stream');

var sources = {}, destinations = {}, encoders = {}, pipes = {}, callbacks = [];

var Streamcore = function(l) {

    logger = l;
    ffmpeg = require('../streaming/ffmpeg.js')(l);

    ffmpeg.check(function (available) {
        callbacks.forEach(function (cb) {
            cb(available);
        });
    });

    Streamcore.playRecording = ffmpeg.playRecording;
    Streamcore.getEncoder = ffmpeg.getEncoder;

    Streamcore.initialised = true;
};

Streamcore.initialised = false;

Streamcore.ffmpeg = function (cb) {
    if (ffmpeg.available !== null) {
        cb(ffmpeg.available);
    } else {
        callbacks.push(cb);
    }
};

Streamcore.getThumbnail = function (id, cb)
{
    var cameras = cfg.get('cameras', {});

    var camera = cameras[id];

    if (!camera) {
        return false;
    }

    getThumbnailImage(id, function (image) {

        if (!image) {
            return false;
        }

        saveImageDimensions(camera, image);

        Streamcore.makeThumbnail(image, camera, cb);
    });
};

Streamcore.runStream = function (cameraid, camera, options, destination, key, thumbnail)
{
    if (!ffmpeg.available && options.encoder && options.encoder !== 'jpeg') {
        logger.error('Video encoding requires ffmpeg to be installed');
        return false;
    }

    destinations[key] = destination;

    destination.on('pipe', function () {
        if (!pipes[cameraid]) {
            pipes[cameraid] = {};
        }
        if (!pipes[cameraid][options.framerate]) {
            pipes[cameraid][options.framerate] = 0;
        }
        pipes[cameraid][options.framerate]++;
    });

    destination.on('unpipe', function () {
        if (--pipes[cameraid][options.framerate] === 0) {
            if (sources[cameraid][options.framerate]) {
                sources[cameraid][options.framerate].end();
            }
        }
    });

    destination.on('finish', function () {
        delete destinations[key];
    });

    var pipeSource = function (source) {

        if (thumbnail) {
            Streamcore.thumbnail(source, camera, thumbnail);
        }

        if (ffmpeg.available && (options.width > 0 || options.height > 0 || camera.rotate || (options.encoder && options.encoder !== 'jpeg'))) {

            var encoder = ffmpeg.getEncoder(options, camera.rotate);

            if (!encoder) {
                return false;
            }

            encoders[key] = encoder;

            source = source.pipe(encoders[key]);

            encoders[key].on('finish', function () {
                delete encoders[key];
            });
        }

        source.pipe(destination);
    };

    if (sources[cameraid] && sources[cameraid][options.framerate] && sources[cameraid][options.framerate].readable) {
        pipeSource(sources[cameraid][options.framerate]);
    } else {

        getSourceStream(camera, options, function (source) {

            if (!source) {
                return false;
            }

            source.setMaxListeners(100);

            if (!sources[cameraid]) {
                sources[cameraid] = {};
            }

            sources[cameraid][options.framerate] = source;

            source.on('end', function() {
                delete sources[cameraid][options.framerate];
            });

            pipeSource(source);
        });
    }
};

Streamcore.thumbnail = function (source, camera, thumbnail) {

    var withCallback = new FirstN(1, function (err, resp) {

        source.unpipe(withCallback);

        if (err) {
            console.log(err);
            return;
        }

        fs.writeFile(thumbnail, resp[0]);
    });

    if (ffmpeg.available) {

        var options = {
            width: -1,
            height: 120
        };

        var scaler = ffmpeg.getEncoder(options, camera.rotate);

        if (scaler) {
            source = source.pipe(scaler);
        }
    }

    source.pipe(withCallback);
}

Streamcore.stopStreaming = function (cameraid, options, key) {

    if (encoders[key]) {
        sources[cameraid][options.framerate].unpipe(encoders[key]);
        encoders[key].unpipe(destinations[key]);
        encoders[key].end();
    } else if (sources[cameraid] && sources[cameraid][options.framerate]) {
        sources[cameraid][options.framerate].unpipe(destinations[key]);
    }

    if (destinations[key]) {
        destinations[key].end();
    }
};

Streamcore.makeThumbnail = function (image, camera, cb) {

    var options = {
        width: -1,
        height: 120
    };

    if (ffmpeg.available) {

        var scaler = ffmpeg.getEncoder(options, camera.rotate);

        if (!scaler) {
            cb(false);
            return;
        }

        scaler.once('data', function (thumbnail) {
            cb(thumbnail);
        });

        scaler.write(image);
        scaler.end();

    } else {
        cb(image);
    }
};

Streamcore.finalize = function (filename, cb) {
    ffmpeg.finalize(filename, cb);
};

function getSourceStream(camera, options, cb)
{
    var method;

    if (camera.mjpeg) {
        method = require('../streaming/mjpeg.js');
    } else if (camera.rtsp && ffmpeg.available) {
        method = require('../streaming/rtsp.js');
    } else if (camera.screencapture) {
        method = require('../streaming/screencapture.js');
    } else if (camera.local) {
        method = require('../streaming/local.js');
    } else if (camera.snapshot) {
        method = require('../streaming/snapshot.js');
    } else {
        logger.error('No valid source found for camera', camera.name);
        cb(false);
    }

    method.stream(logger, camera, options, cb);
}

function saveImageDimensions(camera, image)
{
    var size = require('jpeg-size');

    var s = size(image);

    if (s) {

        if (camera.width !== s.width || camera.height !== s.height) {

            var cameras = cfg.get('cameras', {});

            cameras[camera.id].width = s.width;
            cameras[camera.id].height = s.height;

            cfg.set('cameras', cameras);
        }
    }
}

function getThumbnailImage(id, cb)
{
    var method;

    var cameras = cfg.get('cameras', {});

    var camera = cameras[id];

    if (!camera) {
        logger.debug('Unknown camera', id);
        if (typeof cb === 'function') {
            cb();
        }
        return;
    }

    if (camera.snapshot) {
        method = require('../streaming/snapshot.js');
    } else if (camera.mjpeg) {
        method = require('../streaming/mjpeg.js');
    } else if (camera.rtsp && ffmpeg.available) {
        method = require('../streaming/rtsp.js');
    } else if (camera.screencapture) {
        method = require('../streaming/screencapture.js');
    } else if (camera.local) {
        method = require('../streaming/local.js');
    } else {
        logger.error('No valid source found for camera', camera.name);
        cb(false);
    }

    method.snapshot(logger, camera, cb);
}

module.exports = Streamcore;

